@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (c) 1998-2003 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (c) 1998-2006 Michal Trojnara <Michal.Trojnara@mirt.net>
  *                 All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -23,36 +23,46 @@
 
 #include "common.h"
 
+/**************************************** Network data structure */
+
+#define MAX_HOSTS 16
+
+typedef union sockaddr_union {
+    struct sockaddr sa;
+    struct sockaddr_in in;
+#if defined(USE_IPv6)
+    struct sockaddr_in6 in6;
+#endif
+} SOCKADDR_UNION;
+
+typedef struct sockaddr_list {      /* list of addresses */
+    SOCKADDR_UNION addr[MAX_HOSTS]; /* the list of addresses */
+    u16 cur;                        /* current address for round-robin */
+    u16 num;                        /* how many addresses are used */
+} SOCKADDR_LIST;
+
+#ifdef __INNOTEK_LIBC__
+#define socklen_t    __socklen_t
+#define strcasecmp   stricmp
+#define strncasecmp  strnicmp
+#define NI_NUMERICHOST 1
+#define NI_NUMERICSERV 2
+#endif
+
+
 /**************************************** Prototypes for stunnel.c */
 
-extern int num_clients;
+extern volatile int num_clients;
 
-void main_initialize(char *);
+void main_initialize(char *, char *);
 void main_execute(void);
-void ioerror(char *);
-void sockerror(char *);
-void log_error(int, int, char *);
-char *my_strerror(int);
-int set_socket_options(int, int);
-char *stunnel_info(void);
-int alloc_fd(int);
-char *safe_ntoa(char *, struct in_addr);
-
-/**************************************** Prototypes for ssl.c */
-
-void context_init(void);
-void context_free(void);
-void sslerror(char *);
+void stunnel_info(int);
 
 /**************************************** Prototypes for log.c */
 
 void log_open(void);
 void log_close(void);
-#if defined (USE_WIN32) || defined (__vms)
-/* This conflicts with the "double log (double __x)" routine from math.h */
-#define log stunnel_log
-#endif
-void log(int, const char *, ...)
+void s_log(int, const char *, ...)
 #ifdef __GNUC__
     __attribute__ ((format (printf, 2, 3)));
 #else
@@ -64,20 +74,10 @@ void log_raw(const char *, ...)
 #else
     ;
 #endif
-    
-/**************************************** Prototypes for sthreads.c */
-
-typedef enum {
-    CRIT_KEYGEN, CRIT_LIBWRAP, CRIT_NTOA, CRIT_CLIENTS, CRIT_WIN_LOG,
-    CRIT_SECTIONS
-} section_code;
-
-void enter_critical_section(section_code);
-void leave_critical_section(section_code);
-void sthreads_init(void);
-unsigned long stunnel_process_id(void);
-unsigned long stunnel_thread_id(void);
-int create_client(int, int, void *, void *(*)(void *));
+void ioerror(const char *);
+void sockerror(const char *);
+void log_error(int, int, const char *);
+char *my_strerror(int);
 
 /**************************************** Prototypes for pty.c */
 /* Based on Public Domain code by Tatu Ylonen <ylo@cs.hut.fi>  */
@@ -88,22 +88,31 @@ void pty_release(char *);
 void pty_make_controlling_tty(int *, char *);
 #endif
 
+/**************************************** Prototypes for ssl.c */
+
+typedef enum {
+    COMP_NONE, COMP_ZLIB, COMP_RLE
+} COMP_TYPE;
+
+extern int cli_index;
+
+void ssl_init(void);
+void ssl_configure(void);
+#ifdef HAVE_OSSL_ENGINE_H
+void open_engine(const char *);
+void ctrl_engine(const char *, const char *);
+void close_engine(void);
+ENGINE *get_engine(int);
+#endif
+
 /**************************************** Prototypes for options.c */
 
 typedef struct {
         /* some data for SSL initialization in ssl.c */
-    char *ca_dir;                              /* directory for hashed certs */
-    char *ca_file;                       /* file containing bunches of certs */
-    char *cipher_list;
-    char *cert;                                             /* cert filename */
+    COMP_TYPE compression;                               /* compression type */
     char *egd_sock;                       /* entropy gathering daemon socket */
-    char *key;                               /* pem (priv key/cert) filename */
     char *rand_file;                                /* file with random data */
     int random_bytes;                       /* how many random bytes to read */
-    long session_timeout;
-    int verify_level;
-    int verify_use_only_my;
-    long ssl_options;
 
         /* some global data for stunnel.c */
 #ifndef USE_WIN32
@@ -117,9 +126,8 @@ typedef struct {
 #endif
 
         /* Win32 specific data for gui.c */
-#ifdef USE_WIN32
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
     char *win32_service;
-    char *win32_name;
 #endif
 
         /* logging-support data for log.c */
@@ -131,48 +139,81 @@ typedef struct {
 
         /* on/off switches */
     struct {
-        unsigned int cert:1;
-        unsigned int client:1;
         unsigned int foreground:1;
         unsigned int syslog:1;                              /* log to syslog */
-        unsigned int rand_write;                      /* overwrite rand_file */
+        unsigned int rand_write:1;                    /* overwrite rand_file */
+#ifdef USE_WIN32
+        unsigned int taskbar:1;                   /* enable the taskbar icon */
+#endif
     } option;
 } GLOBAL_OPTIONS;
 
 extern GLOBAL_OPTIONS options;
 
 typedef struct local_options {
+    SSL_CTX *ctx;                                            /*  SSL context */
+    X509_STORE *revocation_store;             /* cert store for CRL checking */
+#ifdef HAVE_OSSL_ENGINE_H
+    ENGINE *engine;                        /* engine to read the private key */
+#endif
     struct local_options *next;            /* next node in the services list */
+    char *servname;        /* service name for logging & permission checking */
+    SSL_SESSION *session;                           /* jecently used session */
+    char local_address[IPLEN];             /* dotted-decimal address to bind */
 
-    char local_address[16]; /* Dotted-decimal address to bind */
-
-        /* name of service */
-    char *servname;         /* service name for loggin & permission checking */
+        /* service-specific data for ctx.c */
+    char *ca_dir;                              /* directory for hashed certs */
+    char *ca_file;                       /* file containing bunches of certs */
+    char *crl_dir;                              /* directory for hashed CRLs */
+    char *crl_file;                       /* file containing bunches of CRLs */
+    char *cipher_list;
+    char *cert;                                             /* cert filename */
+    char *key;                               /* pem (priv key/cert) filename */
+    long session_timeout;
+    int verify_level;
+    int verify_use_only_my;
+    long ssl_options;
+#if SSLEAY_VERSION_NUMBER >= 0x00907000L
+    SOCKADDR_LIST ocsp_addr;
+    char *ocsp_path;
+    unsigned long ocsp_flags;
+#endif /* OpenSSL-0.9.7 */
+    SSL_METHOD *(*client_method)(void), *(*server_method)(void);
 
         /* service-specific data for client.c */
     int fd;        /* file descriptor accepting connections for this service */
-    unsigned short localport, remoteport;
     char *execname, **execargs; /* program name and arguments for local mode */
-    u32 *localnames, *remotenames;
-    u32 *local_ip;
+    SOCKADDR_LIST local_addr, remote_addr;
+    SOCKADDR_LIST source_addr;
     char *username;
     char *remote_address;
     int timeout_busy; /* Maximum waiting for data time */
-    int timeout_idle; /* Maximum idle connection time */
     int timeout_close; /* Maximum close_notify time */
+    int timeout_connect; /* Maximum connect() time */
+    int timeout_idle; /* Maximum idle connection time */
 
         /* protocol name for protocol.c */
     char *protocol;
+    char *protocol_host;
+    char *protocol_username;
+    char *protocol_password;
+    char *protocol_authentication;
 
         /* on/off switches */
     struct {
+        unsigned int cert:1;
+        unsigned int client:1;
         unsigned int delayed_lookup:1;
         unsigned int accept:1;
         unsigned int remote:1;
+        unsigned int retry:1; /* loop remote+program */
 #ifndef USE_WIN32
         unsigned int program:1;
         unsigned int pty:1;
         unsigned int transparent:1;
+#endif
+#if SSLEAY_VERSION_NUMBER >= 0x00907000L
+        unsigned int ocsp:1;
 #endif
     } option;
 } LOCAL_OPTIONS;
@@ -199,8 +240,46 @@ typedef struct {
     OPT_UNION *opt_val[3];
 } SOCK_OPT;
 
-void parse_config(char *);
-int name2nums(char *, char *, u32 **, u_short *);
+void parse_config(char *, char *);
+
+/**************************************** Prototypes for ctx.c */
+
+void context_init(LOCAL_OPTIONS *);
+void sslerror(char *);
+
+/**************************************** Prototypes for verify.c */
+
+void verify_init(LOCAL_OPTIONS *);
+
+/**************************************** Prototypes for network.c */
+
+#ifdef USE_POLL
+#define MAX_FD 256
+#endif
+
+typedef struct {
+#ifdef USE_POLL
+    struct pollfd ufds[MAX_FD];
+    unsigned int nfds;
+#else
+    fd_set irfds, iwfds, orfds, owfds;
+    int max;
+#endif
+} s_poll_set;
+
+void s_poll_zero(s_poll_set *);
+void s_poll_add(s_poll_set *, int, int, int);
+int s_poll_canread(s_poll_set *, int);
+int s_poll_canwrite(s_poll_set *, int);
+int s_poll_wait(s_poll_set *, int, int);
+
+#ifndef USE_WIN32
+int signal_pipe_init(void);
+void child_status(void);  /* dead libwrap or 'exec' process detected */
+#endif
+int set_socket_options(int, int);
+int alloc_fd(int);
+void setnonblock(int, unsigned long);
 
 /**************************************** Prototypes for client.c */
 
@@ -213,14 +292,16 @@ typedef struct {
 
 typedef struct {
     LOCAL_OPTIONS *opt;
-    char accepting_address[16], connecting_address[16]; /* Dotted-decimal */
-    struct sockaddr_in addr; /* Local address */
+    char accepted_address[IPLEN]; /* text */
+    SOCKADDR_LIST peer_addr; /* Peer address */
     FD local_rfd, local_wfd; /* Read and write local descriptors */
-    FD remote_fd; /* Remote descriptor */
+    FD remote_fd; /* Remote file descriptor */
     SSL *ssl; /* SSL Connection */
-    int bind_ip; /* IP for explicit local bind or transparent proxy */
+    SOCKADDR_LIST bind_addr;
+        /* IP for explicit local bind or transparent proxy */
     unsigned long pid; /* PID of local process */
-    u32 *resolved_addresses; /* List of IP addresses for delayed lookup */
+    int fd; /* Temporary file descriptor */
+    jmp_buf err;
 
     char sock_buff[BUFFSIZE]; /* Socket read buffer */
     char ssl_buff[BUFFSIZE]; /* SSL read buffer */
@@ -228,6 +309,7 @@ typedef struct {
     FD *sock_rfd, *sock_wfd; /* Read and write socket descriptors */
     FD *ssl_rfd, *ssl_wfd; /* Read and write SSL descriptors */
     int sock_bytes, ssl_bytes; /* Bytes written to socket and ssl */
+    s_poll_set fds; /* File descriptors */
 } CLI;
 
 extern int max_clients;
@@ -235,27 +317,16 @@ extern int max_clients;
 extern int max_fds;
 #endif
 
-#define sock_rd (c->sock_rfd->rd)
-#define sock_wr (c->sock_wfd->wr)
-#define ssl_rd (c->ssl_rfd->rd)
-#define ssl_wr (c->ssl_wfd->wr)
-
 void *alloc_client_session(LOCAL_OPTIONS *, int, int);
 void *client(void *);
+int connect_wait(CLI *);
 
-/**************************************** Prototype for protocol.c */
-
-int negotiate(CLI *c);
-
-/**************************************** Prototypes for select.c */
-
-int sselect(int, fd_set *, fd_set *, fd_set *, struct timeval *);
-int waitforsocket(int, int, int);
-#ifndef USE_WIN32
-void sselect_init(fd_set *, int *);
-void exec_status(void);
-#endif
+/**************************************** Prototypes for network.c */
 
+void write_blocking(CLI *, int fd, void *, int);
+void read_blocking(CLI *, int fd, void *, int);
+void fdputline(CLI *, int, const char *);
+void fdgetline(CLI *, int, char *);
 /* descriptor versions of fprintf/fscanf */
 int fdprintf(CLI *, int, const char *, ...)
 #ifdef __GNUC__
@@ -270,12 +341,99 @@ int fdscanf(CLI *, int, const char *, char *)
        ;
 #endif
 
+/**************************************** Prototype for protocol.c */
+
+void negotiate(CLI *c);
+
+/**************************************** Prototypes for resolver.c */
+
+int name2addrlist(SOCKADDR_LIST *, char *, char *);
+int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+char *s_ntop(char *, SOCKADDR_UNION *);
+
+/**************************************** Prototypes for sthreads.c */
+
+typedef enum {
+    CRIT_KEYGEN, CRIT_INET, CRIT_CLIENTS, CRIT_WIN_LOG, CRIT_SESSION,
+    CRIT_SECTIONS
+} SECTION_CODE;
+
+void enter_critical_section(SECTION_CODE);
+void leave_critical_section(SECTION_CODE);
+void sthreads_init(void);
+unsigned long stunnel_process_id(void);
+unsigned long stunnel_thread_id(void);
+int create_client(int, int, void *, void *(*)(void *));
+#ifdef USE_UCONTEXT
+typedef struct CONTEXT_STRUCTURE {
+    char stack[STACK_SIZE];
+    unsigned long id;
+    ucontext_t ctx;
+    s_poll_set *fds;
+    int ready; /* number of ready file descriptors */
+    time_t finish; /* when to finish poll() for this context */
+    struct CONTEXT_STRUCTURE *next; /* next context on a list */
+} CONTEXT;
+extern CONTEXT *ready_head, *ready_tail;
+extern CONTEXT *waiting_head, *waiting_tail;
+#endif
+#ifdef _WIN32_WCE
+int _beginthread(void (*)(void *), int, void *);
+void _endthread(void);
+#endif
+#ifdef DEBUG_STACK_SIZE
+void stack_info(int);
+#endif
+
 /**************************************** Prototypes for gui.c */
 
+typedef struct {
+    LOCAL_OPTIONS *section;
+    char pass[PEM_BUFSIZE];
+} UI_DATA;
+
 #ifdef USE_WIN32
 void win_log(char *);
 void exit_stunnel(int);
-int pem_passwd_cb(char *, int, int, void *);
+int passwd_cb(char *, int, int, void *);
+#ifdef HAVE_OSSL_ENGINE_H
+int pin_cb(UI *, UI_STRING *);
+#endif
+
+#ifndef _WIN32_WCE
+typedef int (CALLBACK * GETADDRINFO) (const char *,
+    const char *, const struct addrinfo *, struct addrinfo **);
+typedef void (CALLBACK * FREEADDRINFO) (struct addrinfo FAR *);
+typedef int (CALLBACK * GETNAMEINFO) (const struct sockaddr *, socklen_t,
+    char *, size_t, char *, size_t, int);
+extern GETADDRINFO s_getaddrinfo;
+extern FREEADDRINFO s_freeaddrinfo;
+extern GETNAMEINFO s_getnameinfo;
+#endif /* ! _WIN32_WCE */
+#endif /* USE_WIN32 */
+
+/**************************************** Prototypes for file.c */
+
+typedef struct disk_file {
+#ifdef USE_WIN32
+    HANDLE fh;
+#else
+    int fd;
+#endif
+    /* the inteface is prepared to easily implement buffering if needed */
+} DISK_FILE;
+
+#ifndef USE_WIN32
+DISK_FILE *file_fdopen(int);
+#endif
+DISK_FILE *file_open(char *, int);
+void file_close(DISK_FILE *);
+int file_getline(DISK_FILE *, char *, int);
+int file_putline(DISK_FILE *, char *);
+
+#ifdef USE_WIN32
+LPTSTR str2tstr(const LPSTR);
+LPSTR tstr2str(const LPTSTR);
 #endif
 
 #endif /* defined PROTOTYPES_H */
